'From Squeak3.5 of ''11 April 2003'' [latest update: #5180] on 28 June 2003 at 12:42:28 am'!
Object subclass: #ContextTag
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!ContextTag commentStamp: 'ajh 1/27/2003 19:50' prior: 0!
Instead of referring to a specific context of a specific context stack, use 'context contextTag' to refer to that context and any of its copies.  Contexts get copied when using continuations, but their contextTags stay the same.  To get a context in the current thread for a contextTag send #context to the contextTag.

BlockClosures hold contextTags for remote returns.  Sending #return: to a contextTag will find the context in the current thread for that contextTag and return from it.
!
]style[(77 7 405)f1,f1i,f1!

Object subclass: #Continuation
	instanceVariableNames: 'suspendedContext '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!Continuation commentStamp: 'ajh 1/27/2003 19:51' prior: 0!
I contain a suspended context chain that gets copied and resumed upon evaluation (ala Scheme).

The following example will execute 'hello' inspect twice, once after normal return and once after continuation evaluation (resumeValue value: nil):

	| resumeValue |
	resumeValue := [:continuation | continuation] callCC.
	'hello' inspect.
	resumeValue ifNotNil: [
		"resumeValue equals the continuation the first time.
		 Now resume the continuation with a resume value of nil"
		resumeValue value: nil].

Thanks to Avi Bryant for explaining continuations and providing the above example.
!
]style[(131 15 73 22 344)f1,f1i,f1,f1i,f1!

ProtoObject subclass: #MessageCatcher
	instanceVariableNames: 'echoToTranscript '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!MessageCatcher commentStamp: '<historical>' prior: 0!
Any message sent to me is returned as a Message object.

"Message catcher" creates an instance of me.
!


!ContextPart methodsFor: 'query' stamp: 'ajh 1/27/2003 19:07'!
asContinuation

	^ Continuation new suspendedContext: self copyStack! !

!ContextPart methodsFor: 'query' stamp: 'ajh 1/21/2003 00:32'!
privContextTag

	^ nil! !


!BlockContext methodsFor: 'scheduling' stamp: 'ajh 1/27/2003 18:51'!
callCC
	"Call with current continuation, ala Scheme.
	Evaluate self against a copy of the sender's call stack, which can be resumed later"

	^ self value: thisContext sender asContinuation! !


!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 10/24/2002 17:43'!
cannotReturn: value
	"The VM tried to return value from the context in the active process that has self as its contextTag, but no such context was found.  Sent from primitive return: failure code.  Also sent directly from the VM when the remoteReturn bytecode fails, in which case the problem block is thisContext client"

	^ BlockCannotReturn new
		result: self
		signal! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 1/27/2003 18:26'!
context
	"Return the context in the active process that has this tag.  Raise an error if none found"

	^ thisContext findContextSuchThat: [:c | c privContextTag == self]! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ar 6/28/2003 00:13'!
findContextSuchThat: aBlock
	^self context findContextSuchThat: aBlock! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 01:04'!
nextHandlerContext

	^ self context nextHandlerContext! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 1/20/2003 23:31'!
printOn: stream

	super printOn: stream.
	stream space; nextPut: $(.
	self identityHash printOn: stream.
	stream nextPut: $).
! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 1/29/2003 13:04'!
restart

	self context restart! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 1/29/2003 13:04'!
restartWithNewReceiver: obj

	self context
		swapReceiver: obj;
		restart! !

!ContextTag methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 12:29'!
return: value
	"Find the context in the active process that has this tag and return value from it.  Execute any ensure:/ifCurtailed: blocks above it first.  If no context is found raise cannot return error."

	| home |
	home _ thisContext sender.
	[home privContextTag == self] whileFalse: [
		home _ home sender.
		home ifNil: [^ self cannotReturn: value].
	].
	home return: value.
! !


!Continuation methodsFor: 'as yet unclassified' stamp: 'ajh 1/27/2003 21:21'!
longPrintOn: stream

	| ctxt |
	super printOn: stream.
	stream cr.
	ctxt _ self suspendedContext.
	[ctxt == nil] whileFalse: [
		stream space.
		ctxt printOn: stream.
		stream cr.
		ctxt _ ctxt sender.
	].
! !

!Continuation methodsFor: 'as yet unclassified' stamp: 'ajh 1/27/2003 21:22'!
suspendedContext

	^ suspendedContext! !

!Continuation methodsFor: 'as yet unclassified' stamp: 'ajh 1/27/2003 19:01'!
suspendedContext: aContext

	suspendedContext _ aContext! !

!Continuation methodsFor: 'as yet unclassified' stamp: 'ajh 2/9/2003 18:22'!
value
	"Copy and resume my thread.  Send this to results of 'Continuation current'"

	thisContext sender terminateTo: thisContext bottom.
	suspendedContext copyStack jump.

"The first line (terminateTo) is necessary for BlockContexts, which if not nilled out refuse to be evaluated again.  You'll find the test cases for Continuation and Amb won't run properly without the termination - at least not without recompiling TestRunner to use BlockClosure instead of BlockContext."
! !

!Continuation methodsFor: 'as yet unclassified' stamp: 'ajh 2/9/2003 18:22'!
value: resumeValue
	"Copy and resume my thread.  Send this to results of #callCC (or #currentDo:).  resumeValue will be the return value of callCC"

	thisContext sender terminateTo: thisContext bottom.
	suspendedContext copyStack
		push: resumeValue;
		jumpTop.

"The first line (terminateTo) is necessary for BlockContexts, which if not nilled out refuse to be evaluated again.  You'll find the test cases for Continuation and Amb won't run properly without the termination - at least not without recompiling TestRunner to use BlockClosure instead of BlockContext."
! !


!Continuation class methodsFor: 'as yet unclassified' stamp: 'ajh 1/27/2003 19:08'!
current

	^ thisContext sender asContinuation! !

!Continuation class methodsFor: 'as yet unclassified' stamp: 'ab 2/4/2003 22:27'!
currentDo: aBlock

	^ aBlock value: thisContext sender asContinuation! !


!Message class methodsFor: 'instance creation' stamp: 'ajh 7/11/2001 12:05'!
catcher

	^ MessageCatcher new! !


!MessageCatcher methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 12:24'!
doesNotUnderstand: aMessage

	echoToTranscript == true ifTrue: [Transcript show: aMessage printString; cr].
	^ aMessage! !

!MessageCatcher methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 12:23'!
privEchoToTranscript

	echoToTranscript _ true! !


!MethodContext methodsFor: 'closure support' stamp: 'ar 6/28/2003 00:32'!
privContextTag
	"See contextTag"

	^ receiverMap! !

!MethodContext methodsFor: 'closure support' stamp: 'ar 6/28/2003 00:32'!
contextTag
	"ContextTags are preferred to actual contexts since they are not tied to a specific process, allowing block returns to work in copies of processes (continuations).  ContextTags will always look in the active process for the context with its tag.  receiverMap is used to hold the context's contextTag, since it is unused and we can't add more vars without changing the VM."

	^ receiverMap ifNil: [receiverMap _ ContextTag new]! !

